﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Reflection;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace AsyncCalcPiPassingState
{

    partial class AsyncCalcPiForm : Form
    {

        CancellationTokenSource tokenSource;
        CancellationToken token;
        Task<TimeSpan> piTask = null;
        ManualResetEvent pauseEvent = new ManualResetEvent(true);
        bool bPaused = false;
        Action<CalcPiUserState> ActionShowProgressState;

        public AsyncCalcPiForm()
        {
            InitializeComponent();
            tokenSource = new CancellationTokenSource();
            token = tokenSource.Token;
            SetUIReady();
            ActionShowProgressState = ShowProgressState;
        }

        void ShowProgressState(CalcPiUserState state)
        {
            try
            {
                // Make sure we're on the UI thread
                Debug.Assert(this.InvokeRequired == false);

                // Display progress in UI
                this.resultsTextBox.Text = state.Pi;
                this.calcToolStripProgressBar.Maximum = state.TotalDigits;
                this.calcToolStripProgressBar.Value = state.DigitsSoFar;
                state.CalculateDigits = (int)numericUpDownChunks.Value;
                if (checkBoxThrowError.Checked)
                {
                    state.CalculateDigits = -1;
                }

            } catch (Exception e)
            {
                Console.WriteLine("{0} in ShowProgressState", e.Message);
                throw e;
            }
        }

        TimeSpan TimePiThread(
            int digits, IProgress<CalcPiUserState> progress, CancellationToken token)
        {
            // Track start time
            DateTime start = DateTime.Now;

            CalcPiWithProgressAndCancellation(digits, progress, token);

            // Return elapsed time
            DateTime end = DateTime.Now;
            TimeSpan elapsed = end - start;
            return elapsed;
        }

        void ProgressTestPauseCancelError(
            IProgress<CalcPiUserState> progress, CancellationToken token, CalcPiUserState state)
        {
            // Report progress
            //if (progress != null)
            //    progress.Report(state);
            this.Invoke(ActionShowProgressState, state);
            if (state.CalculateDigits < 0)
            {
                throw new Exception("Oooops!");
            }
            token.ThrowIfCancellationRequested();
            pauseEvent.WaitOne();
        }

        void CalcPiWithProgressAndCancellation(
            int digits, IProgress<CalcPiUserState> progress, CancellationToken token)
        {
            StringBuilder pi = new StringBuilder("3", digits + 2);

            CalcPiUserState state =
                new CalcPiUserState(pi.ToString(), digits, 0, 9);

            //Report initial progress
            ProgressTestPauseCancelError(progress, token, state);

            if (digits > 0)
            {
                pi.Append(".");

                for (int i = 0; i < digits; i += state.CalculateDigits)
                {
                    int digitCountTotal = 0;
                    for (int j = 0; j < state.CalculateDigits && state.DigitsSoFar + digitCountTotal < digits; j += 9)
                    {
                        int calcDigits = NineDigitsOfPi.StartingAt(i + j + 1);
                        int digitCount = Math.Min(digits - j - i, 9);
                        digitCountTotal += digitCount;
                        string ds = string.Format("{0:D9}", calcDigits);
                        pi.Append(ds.Substring(0, digitCount));
                    }

                    // Report continuing progress
                    state = new CalcPiUserState(pi.ToString(), digits, i + digitCountTotal, state.CalculateDigits);
                    ProgressTestPauseCancelError(progress, token, state);
                }
            }
        }

        void SetUIBusy()
        {
            // Set calculating UI
            this.calcButton.Text = "Cancel";
            this.calcToolStripProgressBar.Visible = true;
            this.calcToolStripStatusLabel.Text = "Calculating...";
            this.buttonPause.Enabled = true;
        }

        void SetUIReady()
        {
            // Reset progress UI
            this.calcButton.Text = "Calculate";
            this.calcButton.Enabled = true;
            this.calcToolStripStatusLabel.Text = "Ready";
            this.calcToolStripProgressBar.Visible = false;
            this.buttonPause.Enabled = false;
        }

        void ResetTaskAndCancellation()
        {
            piTask = null;
            //A token source cannot be reused. Create a new one.
            tokenSource = new CancellationTokenSource();
            token = tokenSource.Token;
        }

        private TimeSpan WorkCompleted(Task<TimeSpan> task)
        {
            return new TimeSpan(0, 0, 10);
        }

        async
        Task StartWork()
        {
            SetUIBusy();

            //Any method that is an Action, has one parameter of the gerenic type and returns void.
            Action<CalcPiUserState> actionProgress = ShowProgressState;
            //The Action and the Progress must have the same generic type.
            IProgress<CalcPiUserState> progress = new Progress<CalcPiUserState>(actionProgress);

            TimeSpan elapsed = default(TimeSpan);

            try
            {
                //lambda function is required here, since constructor expects Action, that has no parameters.
                piTask = new Task<TimeSpan>(() =>
                 TimePiThread(
                    (int)this.decimalPlacesNumericUpDown.Value, progress, tokenSource.Token)
                 , tokenSource.Token);
                piTask.Start();
                elapsed = await piTask;
                MessageBox.Show("Elapsed: " + elapsed.ToString());
            }
            catch (OperationCanceledException oce)
            {
                Console.WriteLine("*****************************" + oce.Message);
                this.resultsTextBox.Text = "Cancelled";
            }
            catch (Exception e)
            {
                Console.WriteLine("************Exception*****************" + e.Message);
                if (piTask.Exception != null)
                {
                    resultsTextBox.AppendText(String.Format("\r\nFaulted\r\n{0}", piTask.Exception.ToString()));
                }
                else
                {
                    throw;
                }
            }

            if (piTask != null)
            {
                resultsTextBox.AppendText(String.Format("\r\ncompleted = {0}; canceled = {1}; faulted = {2}.",
                    piTask.IsCompleted, piTask.IsCanceled, piTask.IsFaulted));
            }
            SetUIReady();
            ResetTaskAndCancellation();
            if (m_bClosing)
            {
                MessageBox.Show("Closing after thread was cancelled");
                this.Close();
                m_bClosing = false;
            }
        }

        async void calcButton_Click(object sender, EventArgs e)
        {
            // Don't process if cancel request pending
            // Should not happen, since button will be disabled
            if (this.tokenSource.IsCancellationRequested) { return; }

            if (piTask != null && !piTask.IsCompleted)
            {
                this.calcButton.Enabled = false;
                if (bPaused) pauseEvent.Set();
                tokenSource.Cancel();
            }
            else
            {
                SetUIBusy();
                await StartWork();
            }
        }

        bool m_bClosing = false;
        private void AsyncCalcPiForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (checkBoxStopOnClose.Checked && piTask != null && !piTask.IsCompleted)
            {
                tokenSource.Cancel();
                m_bClosing = true;
                e.Cancel = true;
            }
        }

        private void buttonPause_Click(object sender, EventArgs e)
        {
            bPaused = !bPaused;
            if (bPaused)
            {
                buttonPause.Text = "Continue";
                pauseEvent.Reset();
            } else
            {
                buttonPause.Text = "Pause";
                pauseEvent.Set();
            }
        }
    }
}
